/**
 * Written by Cédric Tabin
 * http://www.astorm.ch/
 * Version 1.0 - 08.08.2009
 * 
 * Code under licence GPLv2
 * Geany - http://www.geany.org/
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//defines
#define PRETTY_PRINTING_SUCCESS 0
#define MALLOC_ERROR -1
#define INVALID_CHAR_ERROR 1
#define EMPTY_XML 2
#define NOT_SUPPORTED_YET 3

//some simple types
#define boolean int
#define true 1
#define false 0

//======================= PARAMETERS - CONSTANTS =======================================================

//TODO put all that constants in parameters or external file
//TODO also put this struct in a .h file
typedef struct 
{
      char indentChar;                                                                           //char used for indentation
      int indentLength;                                                                          //number of char to use for indentation (by default 2 spaces)
      boolean oneLineText;                                                                       //text is put on one line  
      boolean inlineText;                                                                        //if possible text are inline (no return after the opening node and before closing node)
      boolean oneLineComment;                                                                    //comments are put on one line
      boolean inlineComment;                                                                     //if possible comments are inline (no return after the opening node and before closing node)
      boolean oneLineCdata;                                                                      //cdata are put on one line
      boolean inlineCdata;                                                                       //if possible cdata are inline (no return after the opening node and before closing node)
      boolean emptyNodeStripping;                                                                //the empty nodes such <node></node> are set to <node/>
      boolean emptyNodeStrippingSpace;                                                           //put a space before the '/>' when a node is stripped
      boolean forceEmptyNodeSplit;                                                               //force an empty node to be splitted : <node /> becomes <node></node> (only if emptyNodeStripping = false)
} PrettyPrintingOptions;

//======================= PRETTY_PRINTING - DO NOT CROSS ===============================================

//TODO create a .h file for public method (actually just processXMLPrettyPrinting and createDefaultPrettyPrintingOptions)

//xml pretty printing functions
int processXMLPrettyPrinting(char** xml, int* length, PrettyPrintingOptions* ppOptions);         //process the pretty-printing on a valid xml string (no check done !!!). The ppOptions ARE NOT FREE-ED after processing. The method returns 0 if the pretty-printing has been done.
PrettyPrintingOptions* createDefaultPrettyPrintingOptions();                                     //creates a default PrettyPrintingOptions object
void putCharInBuffer(char charToAdd);                                                            //put a char into the new char buffer
void putCharsInBuffer(char* charsToAdd);                                                         //put the chars into the new char buffer
void putNextCharsInBuffer(int nbChars);                                                          //put the next nbChars of the input buffer into the new buffer
int readWhites();                                                                                //read the next whites into the input buffer
char readNextChar();                                                                             //read the next char into the input buffer;
char getNextChar();                                                                              //returns the next char but do not increase the input buffer index (use readNextChar for that)
char getPreviousInsertedChar();                                                                  //returns the last inserted char into the new buffer
boolean isWhite(char c);                                                                         //check if the specified char is a white
boolean isLineBreak(char c);                                                                     //check if the specified char is a new line
int putNewLine();                                                                                //put a new line into the new char buffer with the correct number of whites (indentation)
boolean isInlineNodeAllowed();                                                                   //check if it is possible to have an inline node
void resetBackwardIndentation(boolean resetLineBreak);                                           //reset the indentation for the current depth (just reset the index in fact)
                                                             
//specific parsing functions                                 
int processElements();                                                                           //returns the number of elements processed
void processElementAttribute();                                                                  //process on attribute of a node
void processElementAttributes();                                                                 //process all the attributes of a node
void processHeader();                                                                            //process the header <?xml version="..." ?>
void processNode();                                                                              //process an XML node
void processTextNode();                                                                          //process a text node
void processComment();                                                                           //process a comment
void processCDATA();                                                                             //process a CDATA node
void processDoctype();                                                                           //process a DOCTYPE node
void processDoctypeElement();                                                                    //process a DOCTYPE ELEMENT node
                                                             
//debug function                                             
void printDebugStatus();                                                                         //just print some variables into the console for debugging

//============================================ PRIVATE PROPERTIES ======================================

//those are variables that are shared by the functions and
//shouldn't be altered.

int result;                                                                                      //result of the pretty printing
char* xmlPrettyPrinted;                                                                          //new buffer for the formatted XML
int xmlPrettyPrintedLength;                                                                      //buffer size
int xmlPrettyPrintedIndex;                                                                       //buffer index (position of the next char to insert)
char* inputBuffer;                                                                               //input buffer
int inputBufferLength;                                                                           //input buffer size
int inputBufferIndex;                                                                            //input buffer index (position of the next char to read into the input string)
int currentDepth;                                                                                //current depth (for indentation)
char* currentNodeName;                                                                           //current node name
boolean appendIndentation;                                                                       //if the indentation must be added (with a line break before)
boolean lastNodeOpen;                                                                            //defines if the last action was a not opening or not
PrettyPrintingOptions* options;                                                                  //options of PrettyPrinting

//============================================ GENERAL FUNCTIONS =======================================

int processXMLPrettyPrinting(char** buffer, int* length, PrettyPrintingOptions* ppOptions)
{
	//empty buffer, nothing to process
	if (*length == 0) { return EMPTY_XML; }
	if (buffer == NULL || *buffer == NULL) { return EMPTY_XML; }
	
	//initialize the variables
	result = PRETTY_PRINTING_SUCCESS;
	boolean freeOptions = false;
	if (ppOptions == NULL) { ppOptions = createDefaultPrettyPrintingOptions(); freeOptions = true; }
	options = ppOptions;
	
	currentNodeName = NULL;
	appendIndentation = true;
	lastNodeOpen = false;
	xmlPrettyPrintedIndex = 0;
	inputBufferIndex = 0;
	currentDepth = -1;
	
	inputBuffer = *buffer;
	inputBufferLength = *length;
	
	xmlPrettyPrintedLength = *length;
	xmlPrettyPrinted = (char*)malloc(sizeof(char)*(*length));
	if (xmlPrettyPrinted == NULL) { exit(MALLOC_ERROR); }
	
	//go to the first char
	readWhites();

	//process the pretty-printing
	processElements();
	
	//close the buffer
	putCharInBuffer('\0');
	
	//adjust the final size
	xmlPrettyPrinted = realloc(xmlPrettyPrinted, xmlPrettyPrintedIndex);
	if (xmlPrettyPrinted == NULL) { exit(MALLOC_ERROR); }
	
	//freeing the unused values
	if (freeOptions) { free(options); }
	
	//if success, then update the values
	if (result == PRETTY_PRINTING_SUCCESS)
	{
		free(*buffer);
		*buffer = xmlPrettyPrinted;
		*length = xmlPrettyPrintedIndex-1;
	}
	//else clean the other values
	else
	{
		free(xmlPrettyPrinted);
	}
	
	//updating the pointers for the using into the caller function
	xmlPrettyPrinted = NULL; //avoid reference
	inputBuffer = NULL; //avoid reference
	currentNodeName = NULL; //avoid reference
	options = NULL; //avoid reference
	
	//and finally the result
	return result;
}

PrettyPrintingOptions* createDefaultPrettyPrintingOptions()
{
	PrettyPrintingOptions* options = (PrettyPrintingOptions*)malloc(sizeof(PrettyPrintingOptions));
	if (options == NULL) { fprintf(stderr, "createDefaultPrettyPrintingOptions : Unable to allocate memory"); return NULL; }
	
	options->indentChar = ' ';
	options->indentLength = 2;
	options->oneLineText = true;
	options->inlineText = true;
	options->oneLineComment = true;
	options->inlineComment = true;
	options->oneLineCdata = true;
	options->inlineCdata = true;
	options->emptyNodeStripping = true;
	options->emptyNodeStrippingSpace = true;
	options->forceEmptyNodeSplit = false;
	
	return options;
}

void putNextCharsInBuffer(int nbChars)
{
	for (int i=0 ; i<nbChars ; ++i)
	{
		char c = readNextChar();
		putCharInBuffer(c);
	}
}

void putCharInBuffer(char charToAdd)
{
	//check if the buffer is full and reallocation if needed
	if (xmlPrettyPrintedIndex >= xmlPrettyPrintedLength)
	{
		if (charToAdd == '\0') { ++xmlPrettyPrintedLength; }
		else { xmlPrettyPrintedLength += inputBufferLength; }
		xmlPrettyPrinted = (char*)realloc(xmlPrettyPrinted, xmlPrettyPrintedLength);
		if (xmlPrettyPrinted == NULL) { exit(MALLOC_ERROR); }
	}
	
	//putting the char and increase the index for the next one
	xmlPrettyPrinted[xmlPrettyPrintedIndex] = charToAdd;
	++xmlPrettyPrintedIndex;
}

void putCharsInBuffer(char* charsToAdd)
{
	int index = 0;
	while (charsToAdd[index] != '\0')
	{
		putCharInBuffer(charsToAdd[index]);
		++index;
	}
}

char getPreviousInsertedChar()
{
	return xmlPrettyPrinted[xmlPrettyPrintedIndex-1];
}

int putNewLine()
{
	putCharInBuffer('\n');
	int spaces = currentDepth*options->indentLength;
	for(int i=0 ; i<spaces ; ++i)
	{
		putCharInBuffer(options->indentChar);
	}
	
	return spaces;
}

char getNextChar()
{
	return inputBuffer[inputBufferIndex];
}

char readNextChar()
{	
	return inputBuffer[inputBufferIndex++];
}

int readWhites()
{
	int counter = 0;
	while(isWhite(inputBuffer[inputBufferIndex]))
	{
		++counter;
		++inputBufferIndex;
	}
	
	return counter;
}

boolean isWhite(char c)
{
	if (c == ' ') return true;
	if (c == '\t') return true;
	if (c == '\r') return true;
	if (c == '\n') return true;

	return false;
}

boolean isLineBreak(char c)
{
	if (c == '\n') return true;
	if (c == '\r') return true;
	
	return false;
}

boolean isInlineNodeAllowed()
{
	//the last action was not an opening => inline not allowed
	if (!lastNodeOpen) { return false; }
	
	int firstChar = getNextChar(); //should be '<' or we are in a text node
	int secondChar = inputBuffer[inputBufferIndex+1]; //should be '!'
	int thirdChar = inputBuffer[inputBufferIndex+2]; //should be '-' or '['
	
	int index = inputBufferIndex+1;
	if (firstChar == '<')
	{
		//another node is being open ==> no inline !
		if (secondChar != '!') { return false; }
		
		//okay we are in a comment node, so read until it is closed
		
		//select the closing char
		char closingComment = '-';
		if (thirdChar == '[') { closingComment = ']'; }
		
		//read until closing
		char oldChar = ' ';
		index += 3; //that by pass meanless chars
		boolean loop = true;
		while (loop)
		{
			char current = inputBuffer[index];
			if (current == closingComment && oldChar == closingComment) { loop = false; } //end of comment
			oldChar = current;
			++index;
		}
		
		//okay now avoid blanks
		// inputBuffer[index] is now '>'
		++index;
		while (isWhite(inputBuffer[index])) { ++index; }
	}
	else
	{
		//this is a text node. Simply loop to the next '<'
		while (inputBuffer[index] != '<') { ++index; }
	}
	
	//check what do we have now
	char currentChar = inputBuffer[index];
	if (currentChar == '<')
	{
		//check if that is a closing node
		currentChar = inputBuffer[index+1];
		if (currentChar == '/')
		{
			//as we are in a correct XML (so far...), if the node is being
			//directly close, the inline is allowed !!!
			return true;
		}
	}
	
	//inline not allowed...
	return false;
}

void resetBackwardIndentation(boolean resetLineBreak)
{
	xmlPrettyPrintedIndex -= (currentDepth*options->indentLength);
	if (resetLineBreak) { --xmlPrettyPrintedIndex; }
}

//#########################################################################################################################################
//-----------------------------------------------------------------------------------------------------------------------------------------
                                                                                                
//-----------------------------------------------------------------------------------------------------------------------------------------
//=============================================================== NODE FUNCTIONS ==========================================================
//-----------------------------------------------------------------------------------------------------------------------------------------
                                                                                                
//-----------------------------------------------------------------------------------------------------------------------------------------
//#########################################################################################################################################

int processElements()
{
	int counter = 0;
	++currentDepth;
	boolean loop = true;
	while (loop && result == PRETTY_PRINTING_SUCCESS)
	{
		//strip unused whites
		readWhites();
		
		char nextChar = getNextChar();
		if (nextChar == '\0') { return 0; } //no more data to read
		
		//put a new line with indentation
		if (appendIndentation && currentDepth > 0) { putNewLine(); }
		
		//always append indentation (but need to store the state)
		boolean indentBackward = appendIndentation;
		appendIndentation = true; 
		
		//okay what do we have now ?
		if (nextChar != '<')
		{ 
			//a simple text node
			processTextNode(); 
			++counter; 
		} 
		else //some more check are needed
		{
			nextChar = inputBuffer[inputBufferIndex+1];
			if (nextChar == '!') 
			{
				char oneMore = inputBuffer[inputBufferIndex+2];
				if (oneMore == '-') { processComment(); ++counter; } //a comment
				else if (oneMore == '[') { processCDATA(); ++counter; } //cdata
				else if (oneMore == 'D') { processDoctype(); ++counter; } //doctype <!DOCTYPE ... >
				else if (oneMore == 'E') { processDoctypeElement(); ++counter; } //doctype element <!ELEMENT ... >
				else { fprintf(stderr, "processElements : Invalid char '%c' afer '<!'", oneMore); printDebugStatus(); result = INVALID_CHAR_ERROR; }
			} 
			else if (nextChar == '/')
			{ 
				//close a node => stop the loop !!
				loop = false; 
				if (indentBackward) { xmlPrettyPrintedIndex -= options->indentLength; } //INDEX HACKING
			}
			else if (nextChar == '?')
			{
				//this is a header
				processHeader();
			}
			else 
			{
				//a new node is open
				processNode();
				++counter;
			} 
		}
	}
	
	--currentDepth;
	return counter;
}

void processElementAttribute()
{
	//process the attribute name
	char nextChar = readNextChar();
	while (nextChar != '=')
	{
		putCharInBuffer(nextChar);
		nextChar = readNextChar();
	}
	
	putCharInBuffer(nextChar); //that's the '='
	
	//read the simple quote or double quote and put it into the buffer
	char quote = readNextChar();
	putCharInBuffer(quote); 
	
	//process until the last quote
	char value = readNextChar();
	while(value != quote)
	{
		putCharInBuffer(value);
		value = readNextChar();
	}
	
	//simply add the last quote
	putCharInBuffer(quote);
}

void processElementAttributes()
{
	char current = getNextChar(); //should not be a white
	if (isWhite(current)) { fprintf(stderr, "processElementAttributes : first char shouldn't be a white"); printDebugStatus(); result = INVALID_CHAR_ERROR; return; }
	
	boolean loop = true;
	while (loop)
	{
		readWhites(); //strip the whites
		
		char next = getNextChar(); //don't read the last char => will be processed afterwards
		if (next == '/') { loop = false; } /* end of node */
		else if (next == '>') { loop = false; } /* end of tag */
		else if (next == '?') { loop = false; } /* end of header */
		else 
		{ 
			putCharInBuffer(' '); //put only one space to separate attributes
			processElementAttribute(); 
		}
	}
}

void processHeader()
{
	int firstChar = inputBuffer[inputBufferIndex]; //should be '<'
	int secondChar = inputBuffer[inputBufferIndex+1]; //must be '?'
	
	if (firstChar != '<') { fprintf(stderr, "processHeader : first char should be '<' (not '%c')", firstChar); printDebugStatus(); result = INVALID_CHAR_ERROR; return; } /* what ?????? invalid xml !!! */ 
	if (secondChar == '?')
	{ 
		putNextCharsInBuffer(2); //puts the '<' and '?' chars into the new buffer
		
		while(!isWhite(getNextChar())) { putNextCharsInBuffer(1); }
		
		readWhites();
		processElementAttributes(); 
		putNextCharsInBuffer(2); //puts the '?' and '>' chars into the new buffer
		putCharInBuffer('\n');
	}
}

void processNode()
{
	int opening = readNextChar();
	if (opening != '<') { fprintf(stderr, "processNode : The first char should be '<' (not '%c')", opening); printDebugStatus(); result = INVALID_CHAR_ERROR; return; }
	putCharInBuffer(opening);
	
	//read the node name
	int nodeNameLength = 0;
	while (!isWhite(getNextChar()) && getNextChar() != '>' && getNextChar() != '/')
	{
		putNextCharsInBuffer(1);
		++nodeNameLength;
	}

	//store the name
	char* nodeName = (char*)malloc(sizeof(char)*nodeNameLength+1);
	if (nodeName == NULL) { exit(MALLOC_ERROR); }
	nodeName[nodeNameLength] = '\0';
	for (int i=0 ; i<nodeNameLength ; ++i)
	{
		int index = xmlPrettyPrintedIndex-nodeNameLength+i;
		nodeName[i] = xmlPrettyPrinted[index];
	}
	
	currentNodeName = nodeName; //set the name for using in other methods
	lastNodeOpen = true;

	//process the attributes	
	readWhites();
	processElementAttributes();
	
	//process the end of the tag
	int subElementsProcessed = 0;
	char nextChar = getNextChar(); //should be either '/' or '>'
	if (nextChar == '/') //the node is being closed immediatly
	{ 
		//closing node directly
		if (options->emptyNodeStripping || !options->forceEmptyNodeSplit)
		{
			if (options->emptyNodeStrippingSpace) { putCharInBuffer(' '); }
			putNextCharsInBuffer(2); 
		}
		//split the closing nodes
		else
		{
			readNextChar(); //removing '/'
			readNextChar(); //removing '>'
			
			putCharInBuffer('>');
			if (!options->inlineText) { putNewLine(); } //no inline text => new line !
			putCharsInBuffer("</");
			putCharsInBuffer(currentNodeName);
			putCharInBuffer('>');
		}
		
		lastNodeOpen=false; 
		return; 
	}
	else if (nextChar == '>') { putNextCharsInBuffer(1); subElementsProcessed = processElements(true); } //the tag is just closed (maybe some content)
	else { fprintf(stderr, "processNode : Invalid character '%c'", nextChar); printDebugStatus(); result = INVALID_CHAR_ERROR; return; }
	
	//if the code reaches this area, then the processElements has been called and we must
	//close the opening tag
	char closeChar = getNextChar();
	if (closeChar != '<') { fprintf(stderr, "processNode : Invalid character '%c' for closing tag (should be '<')", closeChar); printDebugStatus(); result = INVALID_CHAR_ERROR; return; }
	
	do
	{
		closeChar = readNextChar();
		putCharInBuffer(closeChar);
	}
	while(closeChar != '>');
	
	//there is no elements
	if (subElementsProcessed == 0)
	{
		//the node will be stripped
		if (options->emptyNodeStripping)
		{
			xmlPrettyPrintedIndex -= nodeNameLength+4; //because we have '<nodeName ...></nodeName>'
			resetBackwardIndentation(true);
			
			if (options->emptyNodeStrippingSpace) { putCharInBuffer(' '); }
			putCharsInBuffer("/>");
		}
		//the closing tag will be put on the same line
		else if (options->inlineText)
		{
			xmlPrettyPrintedIndex -= nodeNameLength+3; //because we have '</nodeName>'
			resetBackwardIndentation(true);
			
			//rewrite the node name
			putCharsInBuffer("</");
			putCharsInBuffer(currentNodeName);
			putCharInBuffer('>');
		}
	}
	
	//the node is closed
	lastNodeOpen = false;
	
	//freeeeeeee !!!
	free(nodeName);
	nodeName = NULL;
	currentNodeName = NULL;
}

void processComment()
{
	boolean inlineAllowed = false;
	if (options->inlineComment) { inlineAllowed = isInlineNodeAllowed(); }
	if (inlineAllowed) { resetBackwardIndentation(true); }
	
	putNextCharsInBuffer(4); //add the chars '<!--'
	
	char oldChar = '-';
	boolean loop = true;
	while (loop)
	{
		char nextChar = readNextChar();
		if (oldChar == '-' && nextChar == '-') //the comment is being closed
		{
			loop = false;
		}
		
		if (!isLineBreak(nextChar)) //the comment simply continues
		{
			putCharInBuffer(nextChar);
			oldChar = nextChar;
		}
		else if (!options->oneLineComment) //oh ! there is a line break !
		{
			readWhites(); //strip the whites and new line
			putNewLine(); //put a new indentation line
			oldChar = ' '; //and update the last char
			
			//TODO manage relative spacing
		}
		else //the comments must be inlined
		{
			readWhites(); //strip the whites and add a space if necessary
			if (getPreviousInsertedChar() != ' ') { putCharInBuffer(' '); }
		}
	}
	
	char lastChar = readNextChar(); //should be '>'
	if (lastChar != '>') { fprintf(stderr, "processComment : last char must be '>' (not '%c')", lastChar); printDebugStatus(); result = INVALID_CHAR_ERROR; return; }
	putCharInBuffer(lastChar);
	
	if (inlineAllowed) { appendIndentation = false; }
	
	//there vas no node open
	lastNodeOpen = false;
}

void processTextNode()
{
	//checks if inline is allowed
	boolean inlineTextAllowed = false;
	if (options->inlineText) { inlineTextAllowed = isInlineNodeAllowed(); }
	if (inlineTextAllowed) { resetBackwardIndentation(true); } //remove previous indentation
	
	while(getNextChar() != '<')
	{
		char nextChar = readNextChar();
		if (isLineBreak(nextChar))
		{
			readWhites();
			if (options->oneLineText)
			{ 
				//as we can put text on one line, remove the line break and replace it by a space
				//but only if the previous char wasn't a space
				if (getPreviousInsertedChar() != ' ') { putCharInBuffer(' '); }
			}
			else 
			{
				//put a new line only if the closing tag is not reached
				if (getNextChar() != '<') 
				{	
					putNewLine(); 
				} 
			}
		}
		else
		{
			putCharInBuffer(nextChar);
		}
	}
	
	//remove the indentation for the closing tag
	if (inlineTextAllowed) { appendIndentation = false; }
	
	//there vas no node open
	lastNodeOpen = false;
}

void processCDATA()
{
	boolean inlineAllowed = false;
	if (options->inlineCdata) { inlineAllowed = isInlineNodeAllowed(); }
	if (inlineAllowed) { resetBackwardIndentation(true); }
	
	putNextCharsInBuffer(9); //putting the '<![CDATA[' into the buffer
	
	boolean loop = true;
	char oldChar = '[';
	while(loop)
	{
		char nextChar = readNextChar();
		if (oldChar == ']' && nextChar == ']') { loop = false; } //end of cdata
		
		if (!isLineBreak(nextChar)) //the cdata simply continues
		{
			putCharInBuffer(nextChar);
			oldChar = nextChar;
		}
		else if (!options->oneLineCdata)
		{
			readWhites(); //strip the whites and new line
			putNewLine(); //put a new indentation line
			oldChar = ' '; //and update the last char
			
			//TODO manage relative spacing
		}
		else //cdata are inlined
		{
			readWhites(); //strip the whites and add a space if necessary
			if(getPreviousInsertedChar() != ' ') { putCharInBuffer(' '); }
		}
	}
	
	char lastChar = readNextChar(); //should be '>'
	if (lastChar != '>') { fprintf(stderr, "processCDATA : last char must be '>' (not '%c')", lastChar); printDebugStatus(); result = INVALID_CHAR_ERROR; return; }
	putCharInBuffer(lastChar);
	
	if (inlineAllowed) { appendIndentation = false; }
	
	//there vas no node open
	lastNodeOpen = false;
}

void processDoctype()	
{
	fprintf(stderr, "DOCTYPE is currently not supported by PrettyPrinter\n");
	fflush(stderr);
	
	result = NOT_SUPPORTED_YET;
}

void processDoctypeElement()
{
	fprintf(stderr, "ELEMENT is currently not supported by PrettyPrinter\n");
	fflush(stderr);
	
	result = NOT_SUPPORTED_YET;
}

void printDebugStatus()
{
	fprintf(stderr, "\n===== INPUT =====\n%s\n=================\ninputLength = %d\ninputIndex = %d\noutputLength = %d\noutputIndex = %d\n", 
					inputBuffer, 
					inputBufferLength, 
					inputBufferIndex,
					xmlPrettyPrintedLength,
					xmlPrettyPrintedIndex);
	fflush(stderr);
}
